﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Globalization;
using System.IO;
using System.Linq;
using System.Reflection;
using JetBrains.Annotations;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using OpcLabs.BaseLib.Annotations;
using OpcLabs.BaseLib.Collections.Specialized;
using OpcLabs.BaseLib.Extensions.Internal;
using OpcLabs.BaseLib.Reflection.Extensions;
using OpcLabs.BaseLib.Testing;
using OpcLabs.BaseLib.Utilities;

namespace OpcLabs.UnitTestLib
{
    public class TestSuite
    {
        [NotNull]
        static public string GenerateTestClasses(
            [NotNull] Assembly assembly,
            [NotNull] string baseNamespace,
            [NotNull] string baseTestNamespace,
            bool isPartial = false,
            bool testerInOne = false)
        {
            return GenerateTestClasses(new[] {assembly}, baseNamespace, baseTestNamespace, isPartial, testerInOne);
        }

        // IMPROVE: reimplement using CodeDOM
        [NotNull]
        static public string GenerateTestClasses(
            [NotNull] IEnumerable<Assembly> assemblies,
            [NotNull] string baseNamespace,
            [NotNull] string baseTestNamespace,
            bool isPartial = false,
            bool testerInOne = false)
        {
            using (var memoryStream = new MemoryStream())
            {
                using (var streamWriter = new StreamWriter(memoryStream))
                {
                    streamWriter.WriteLine("using JetBrains.Annotations;");
                    streamWriter.WriteLine("using Microsoft.VisualStudio.TestTools.UnitTesting;");
                    streamWriter.WriteLine("using OpcLabs.BaseLib.Collections.ObjectModel.Extensions;");
                    streamWriter.WriteLine("using OpcLabs.UnitTestLib;");
                    streamWriter.WriteLine("using System.Diagnostics;");
                    streamWriter.WriteLine();

                    // Ordering is useful for meaningful deltas in source control of generated code.
                    IList<Assembly> orderedAssemblies = assemblies
                        .OrderBy(
                            // ReSharper disable once PossibleNullReferenceException
                            assembly => assembly.FullName,
                            StringComparer.Ordinal)
                        .ToList();

                    streamWriter.WriteLine("// Assemblies:");
                    foreach (Assembly assembly in orderedAssemblies)
                        // ReSharper disable once PossibleNullReferenceException
                        streamWriter.WriteLine("//   {0}", assembly.FullName);

                    foreach (Assembly assembly in orderedAssemblies)
                    {
                        Debug.Assert(assembly != null);

                        var assemblyGeneratedTestCategories = new List<string>();
                        IEnumerable<GeneratedTestCategoryAttribute> assemblyGeneratedTestCategoryAttributes = assembly.GetCustomAttributes<GeneratedTestCategoryAttribute>();
                        Debug.Assert(assemblyGeneratedTestCategoryAttributes != null);
                        foreach (GeneratedTestCategoryAttribute assemblyGeneratedTestCategoryAttribute in assemblyGeneratedTestCategoryAttributes)
                        {
                            Debug.Assert(assemblyGeneratedTestCategoryAttribute != null);
                            assemblyGeneratedTestCategories.AddRange(assemblyGeneratedTestCategoryAttribute.TestCategories);
                        }

                        // Ordering is useful for meaningful deltas in source control of generated code.
                        IOrderedEnumerable<Type> orderedtypesUnderTest = assembly.GetTypes().OrderBy(
                            // ReSharper disable once PossibleNullReferenceException
                            type => type.FullName,
                            StringComparer.Ordinal);
                        foreach (Type typeUnderTest in orderedtypesUnderTest)
                        {
                            Debug.Assert(typeUnderTest != null);

                            if (typeUnderTest.IsAbstract)
                                continue;

                            if (typeUnderTest.IsGenericTypeDefinition)
                                continue; // IMPROVE: support this

                            var typeGeneratedTestCategories = new List<string>();
                            IEnumerable<GeneratedTestCategoryAttribute> typeGeneratedTestCategoryAttributes =
                                typeUnderTest.GetCustomAttributes<GeneratedTestCategoryAttribute>();
                            Debug.Assert(typeGeneratedTestCategoryAttributes != null);
                            foreach (GeneratedTestCategoryAttribute typeGeneratedTestCategoryAttribute in typeGeneratedTestCategoryAttributes)
                            {
                                Debug.Assert(typeGeneratedTestCategoryAttribute != null);
                                typeGeneratedTestCategories.AddRange(typeGeneratedTestCategoryAttribute.TestCategories);
                            }

                            // IMPROVE: testers should interpret GeneratedTestCategoryAttribute on member etc. level
                            TesterCollection testers = GetTypeTesters(typeUnderTest);
                            if (testers.Count == 0)
                                continue;

                            bool typeIsObsolete = typeUnderTest.HasCustomAttribute(typeof (ObsoleteAttribute), 
                                inherit: true);

                            string typeNamespace = typeUnderTest.Namespace;
                            Debug.Assert(typeNamespace != null);
                            string testNamespace;
                            if (typeNamespace.StartsWith(baseNamespace)) // IMPROVE: this is not fully correct test!
                                testNamespace = baseTestNamespace + typeNamespace.Substring(baseNamespace.Length);
                            else
                                testNamespace = baseTestNamespace + "." + typeNamespace;

                            string typeName = typeUnderTest.Name;
                            string testClassName = string.Format(CultureInfo.InvariantCulture, "{0}Test", typeName);

                            streamWriter.WriteLine();
                            streamWriter.WriteLine("// ReSharper disable CheckNamespace");
                            streamWriter.WriteLine("namespace {0}", testNamespace);
                            streamWriter.WriteLine("// ReSharper restore CheckNamespace");
                            streamWriter.WriteLine("// ReSharper disable RedundantNameQualifier");
                            streamWriter.WriteLine("{");
                            streamWriter.WriteLine();
                            streamWriter.WriteLine("    [TestClass]");
                            // In case the type under test is inconsistently named, the test class name will likely be
                            // as well, but it makes no sense to report this as a separate warning.
                            streamWriter.WriteLine("    // ReSharper disable InconsistentNaming");
                            if (isPartial)
                                streamWriter.WriteLine("    // ReSharper disable PartialTypeWithSinglePart");
                            streamWriter.WriteLine("    public{0} class {1}", isPartial ? " partial" : "", testClassName);
                            if (isPartial)
                                streamWriter.WriteLine("    // ReSharper restore PartialTypeWithSinglePart");
                            streamWriter.WriteLine("    // ReSharper restore InconsistentNaming");
                            streamWriter.WriteLine("    {");
                            streamWriter.WriteLine("        public {0}()", testClassName);
                            streamWriter.WriteLine("        {");
                            if (typeIsObsolete)
                                streamWriter.WriteLine("#pragma warning disable 0618");
                            streamWriter.WriteLine("            Testers = TestSuite.GetTypeTesters(typeof(global::{0}));",
                                                   typeUnderTest);
                            if (typeIsObsolete)
                                streamWriter.WriteLine("#pragma warning restore 0618");
                            streamWriter.WriteLine("        }");
                            streamWriter.WriteLine();
                            streamWriter.WriteLine("        [TestInitialize]");
                            streamWriter.WriteLine("        public void InitializeTest()");
                            streamWriter.WriteLine("        {");
                            streamWriter.WriteLine("            AssertUtilities.WriteLine(\"Initializing test for {0}...\");", typeUnderTest);
                            streamWriter.WriteLine("        }");
                            streamWriter.WriteLine();
                            streamWriter.WriteLine("        [NotNull]");
                            streamWriter.WriteLine("        public TesterCollection Testers { get; }");

                            // Ordering is useful for meaningful deltas in source control of generated code.
                            IOrderedEnumerable<Tester> orderedTesters = testers.OrderBy(
                                // ReSharper disable once PossibleNullReferenceException
                                tester => tester.Name,
                                StringComparer.Ordinal);
                            foreach (Tester tester in orderedTesters)
                            {
                                Debug.Assert(tester != null);

                                var testerTestCategories = new HashSet<string> { "GeneratedTest" };
                                testerTestCategories.UnionWith(assemblyGeneratedTestCategories);
                                testerTestCategories.UnionWith(typeGeneratedTestCategories);
                                testerTestCategories.UnionWith(tester.TestCategories);

                                streamWriter.WriteLine();
                                if (testerInOne)
                                {
                                    var testCategories = new HashSet<string>(testerTestCategories);
                                    foreach (string testMethodName in tester.TestMethodNames)
                                    {
                                        Debug.Assert(testMethodName != null);
                                        TestMethod testMethod = tester.GetTestMethod(testMethodName);
                                        Debug.Assert(testMethod != null);

                                        foreach (TestCategoryAttribute testCategoryAttribute in testMethod.TestCategoryAttributes)
                                            // ReSharper disable once AssignNullToNotNullAttribute
                                            // ReSharper disable once PossibleNullReferenceException
                                            testCategories.UnionWith(testCategoryAttribute.TestCategories);
                                    }
                                    foreach (string testCategory in testCategories)
                                        streamWriter.WriteLine("        [TestCategory(\"{0}\")]", testCategory);
                                    streamWriter.WriteLine("        [TestMethod]");
                                    streamWriter.WriteLine("        // ReSharper disable once InconsistentNaming");
                                    streamWriter.WriteLine("        public void {0}()", tester.Name);
                                    streamWriter.WriteLine("        {");
                                    streamWriter.WriteLine("            if (!Testers.TryGetValue(\"{0}\", out Tester tester))", tester.Name);
                                    streamWriter.WriteLine("                AssertUtilities.Inconclusive(\"Tester not found.\");");
                                    streamWriter.WriteLine("            Debug.Assert(tester != null);");
                                    streamWriter.WriteLine("            TestUtilities.RunTests(tester.TestMethods);");
                                    streamWriter.WriteLine("        }");
                                }
                                else
                                {
                                    streamWriter.WriteLine("        #region {0} Methods", tester.Name);
                                    // Ordering is useful for meaningful deltas in source control of generated code.
                                    IOrderedEnumerable<string> orderedTestMethodNames = tester.TestMethodNames.OrderBy(
                                        name => name, StringComparer.Ordinal);
                                    foreach (string testMethodName in orderedTestMethodNames)
                                    {
                                        Debug.Assert(testMethodName != null);
                                        TestMethod testMethod = tester.GetTestMethod(testMethodName);
                                        Debug.Assert(testMethod != null);

                                        var testCategories = new HashSet<string>(testerTestCategories);
                                        foreach (TestCategoryAttribute testCategoryAttribute in testMethod.TestCategoryAttributes)
                                            // ReSharper disable once AssignNullToNotNullAttribute
                                            // ReSharper disable once PossibleNullReferenceException
                                            testCategories.UnionWith(testCategoryAttribute.TestCategories);

                                        streamWriter.WriteLine();
                                        streamWriter.WriteLine("        // {0}", testClassName);
                                        foreach (string testCategory in testCategories)
                                            streamWriter.WriteLine("        [TestCategory(\"{0}\")]", testCategory);
                                        streamWriter.WriteLine("        [TestMethod]");
                                        streamWriter.WriteLine("        // ReSharper disable once InconsistentNaming");
                                        streamWriter.WriteLine("        public void {0}_{1}()", tester.Name, testMethodName);
                                        streamWriter.WriteLine("        {");
                                        streamWriter.WriteLine("            FindAndPerformTest(\"{0}\", \"{1}\");",
                                            tester.Name, testMethodName);
                                        streamWriter.WriteLine("        }");
                                    }
                                    streamWriter.WriteLine();
                                    streamWriter.WriteLine("        #endregion");

                                    streamWriter.WriteLine();
                                    streamWriter.WriteLine("        #region Private Methods");
                                    streamWriter.WriteLine();
                                    streamWriter.WriteLine("        // ReSharper disable once UnusedMember.Local");
                                    streamWriter.WriteLine("        private void FindAndPerformTest([NotNull] string testerName, [NotNull] string testMethodName)");
                                    streamWriter.WriteLine("        {");
                                    streamWriter.WriteLine("            if (!Testers.TryGetValue(testerName, out Tester tester))");
                                    streamWriter.WriteLine("                AssertUtilities.Inconclusive(\"Tester not found.\");");
                                    streamWriter.WriteLine("            Debug.Assert(tester != null);");
                                    streamWriter.WriteLine("            if (!tester.FindTestMethod(testMethodName, out TestMethod testMethod))");
                                    streamWriter.WriteLine("                AssertUtilities.Inconclusive(\"Test method not found.\");");
                                    streamWriter.WriteLine("            Debug.Assert(testMethod != null);");
                                    streamWriter.WriteLine("            testMethod.Test();");
                                    streamWriter.WriteLine("        }");
                                    streamWriter.WriteLine();
                                    streamWriter.WriteLine("        #endregion");
                                }
                            }

                            streamWriter.WriteLine("    }");
                            streamWriter.WriteLine();
                            streamWriter.WriteLine("// ReSharper restore RedundantNameQualifier");
                            streamWriter.WriteLine("}");
                        }
                    }
                    streamWriter.Flush();

                    memoryStream.Position = 0;
                    using (var streamReader = new StreamReader(memoryStream))
                        return streamReader.ReadToEnd();
                }
            }
        }

        public struct TypeRecord
        {
            [NotNull]
            public List<TesterAttribute> TesterAttributeList;

            [NotNull]
            public StringObjectDictionary OptionDictionary;
        }

        [NotNull]
        static public TesterCollection GetTypeTesters([NotNull] Type typeUnderTest)
        {
            TypeRecord typeRecord = GetTypeRecord(typeUnderTest);

            var testers = new TesterCollection();
            List<TesterAttribute> testerAttributeList = typeRecord.TesterAttributeList;
            Debug.Assert(testerAttributeList != null);

            StringObjectDictionary optionDictionary = typeRecord.OptionDictionary;
            Debug.Assert(optionDictionary != null);

            foreach (TesterAttribute testerAttribute in testerAttributeList)
            {
                Debug.Assert(testerAttribute != null);

                string testerTypeName = testerAttribute.TesterTypeName;
                Type testerType;
                try
                {
                    testerType = Type.GetType(testerTypeName, throwOnError: true);
                }
                catch (TypeLoadException typeLoadException)
                {
                    throw new ApplicationException(
                        $"Cannot get tester type '{testerTypeName}' for type under test '{typeUnderTest}'.",
                        typeLoadException);
                }
                Debug.Assert(testerType != null);
                var tester = (Tester) ActivatorUtilities.CreateInstance(testerType, BindingFlags.Default, null);
                tester.Parameters = testerAttribute.Parameters;
                tester.OptionDictionary = optionDictionary;
                tester.TypeUnderTest = typeUnderTest;   // has to come AFTER setting the parameters and option dictionary

                // We are allowing duplicite tester specifications but using just one. With tester inheritance, 
                // maintaining non-duplicite testers specifications becomes a burden, 
                // and in the end brings no value.
                if (testers.Remove(tester.Name))
                {
                    // IMPROVE: have some kind of warning?
                }

                testers.Add(tester);
            }
            return testers;
        }

        static public TypeRecord GetTypeRecord([NotNull] Type typeUnderTest)
        {
            Debug.Assert(typeUnderTest != null);

            var optionDictionary = new StringObjectDictionary();
            var testerAttributeList = new List<TesterAttribute>();

            IEnumerable<Type> typeInheritanceChain = typeUnderTest.GetInheritanceChain();
            foreach (Type type in typeInheritanceChain.Reverse())
            {
                Debug.Assert(type != null);

                // ReSharper disable once SuggestUseVarKeywordEvident
                InheritTestersAttribute inheritTestersAttribute =
                    type.GetCustomAttribute<InheritTestersAttribute>();
                bool inherit = true;
                if (inheritTestersAttribute != null)
                    inherit = inheritTestersAttribute.Inherit;

                IEnumerable<TesterAttribute> testerAttributes =
                    type.GetCustomAttributesNotNull<TesterAttribute>(inherit: false);

                if (!inherit)
                    testerAttributeList.Clear();
                testerAttributeList.AddRange(testerAttributes);

                IEnumerable<TestOptionAttribute> testOptionAttributes =
                    type.GetCustomAttributesNotNull<TestOptionAttribute>(inherit: false);
                foreach (TestOptionAttribute testOptionAttribute in testOptionAttributes)
                {
                    Debug.Assert(testOptionAttribute != null);
                    const bool propagate = false;   // IMPROVE: support having it specified in the attribute
                    // ReSharper disable once RedundantLogicalConditionalExpressionOperand
                    if (propagate || (type == typeUnderTest))
                        optionDictionary[testOptionAttribute.Name] = testOptionAttribute.Value;
                }
            }

            TypeRecord typeRecord = new TypeRecord
            {
                TesterAttributeList = testerAttributeList,
                OptionDictionary = optionDictionary
            };
            return typeRecord;
        }
    }
}
