﻿
using System;
using System.Collections.Generic;
using JetBrains.Annotations;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using OpcLabs.BaseLib.Diagnostics;
using OpcLabs.BaseLib.Extensions;
using OpcLabs.BaseLib.Security.Cryptography.PkiCertificates;
using OpcLabs.EasyOpc.UA;
using OpcLabs.EasyOpc.UA.AddressSpace;
using OpcLabs.EasyOpc.UA.Application;
using OpcLabs.EasyOpc.UA.Application.Extensions;
using OpcLabs.EasyOpc.UA.Discovery;
using OpcLabs.EasyOpc.UA.Gds;
using OpcLabs.EasyOpcLiveTest.UA.Environment;
using OpcLabs.UnitTestLib;

namespace OpcLabs.EasyOpcLiveTest.UA.Application
{
    [TestClass]
    public class EasyUAClientApplicationTest
    {
        #region Test Methods

        [TestMethod]
        public void GdsRegistrationLogicTest()
        {
            TestSection.Arrange();
            var gdsEndpointDescriptor = new UAEndpointDescriptor(UATestEnvironment.OpcFoundation_GlobalDiscoveryServer.EndpointDescriptors.OpcTcp)
                { UserIdentity = OpcFoundation_GlobalDiscoveryServer.UserIdentities.Admin };
            AssertUtilities.WriteLine($"gdsEndpointDescriptor: {gdsEndpointDescriptor}");
            TestEnvironment.PrecheckEndpoint(gdsEndpointDescriptor);

            var client = new EasyUAClient();
            IEasyUAClientApplication application = client.GetService<IEasyUAClientApplication>();
            Check.AssertNotNull(application);
            UAApplicationElement applicationElement = application.GetApplicationElement();
            UAGlobalApplicationDiscoveryQuery discoveryQuery = new UAGlobalApplicationDiscoveryQuery(gdsEndpointDescriptor,
                new UAQueryApplicationsFilter { ApplicationUriString = applicationElement.ApplicationUriString });

            TestSection.Act();

            // First unregister
            application.UnregisterFromGds(gdsEndpointDescriptor);
            UADiscoveryElementCollection discoveryElementsAfterFirstUnregister = client.Discover(discoveryQuery);
            Check.AssertNotNull(discoveryElementsAfterFirstUnregister);

            // First register
            UANodeId applicationIdFromFirstRegister = application.RegisterToGds(gdsEndpointDescriptor);
            UADiscoveryElementCollection discoveryElementsAfterFirstRegister = client.Discover(discoveryQuery);
            Check.AssertNotNull(discoveryElementsAfterFirstRegister);

            // Second unregister
            application.UnregisterFromGds(gdsEndpointDescriptor);
            UADiscoveryElementCollection discoveryElementsAfterSecondUnregister = client.Discover(discoveryQuery);
            Check.AssertNotNull(discoveryElementsAfterSecondUnregister);

            // Third unregister
            application.UnregisterFromGds(gdsEndpointDescriptor);
            UADiscoveryElementCollection discoveryElementsAfterThirdUnregister = client.Discover(discoveryQuery);
            Check.AssertNotNull(discoveryElementsAfterThirdUnregister);

            // First update
            UANodeId applicationIdFromFirstUpdate = application.UpdateGdsRegistration(gdsEndpointDescriptor);
            UADiscoveryElementCollection discoveryElementsAfterFirstUpdate = client.Discover(discoveryQuery);
            Check.AssertNotNull(discoveryElementsAfterFirstUpdate);

            // Second register
            UANodeId applicationIdFromSecondRegister = application.RegisterToGds(gdsEndpointDescriptor);
            UADiscoveryElementCollection discoveryElementsAfterSecondRegister = client.Discover(discoveryQuery);
            Check.AssertNotNull(discoveryElementsAfterSecondRegister);

            // Second update
            UANodeId applicationIdFromSecondUpdate = application.UpdateGdsRegistration(gdsEndpointDescriptor);
            UADiscoveryElementCollection discoveryElementsAfterSecondUpdate = client.Discover(discoveryQuery);
            Check.AssertNotNull(discoveryElementsAfterSecondUpdate);

            TestSection.Assert();

            // First unregister
            Assert.IsTrue(discoveryElementsAfterFirstUnregister.Count == 0);

            // First register
            Assert.IsNotNull(applicationIdFromFirstRegister);
            AfterGdsRegistrationAsserts(discoveryElementsAfterFirstRegister, applicationElement);

            // Second unregister
            Assert.IsTrue(discoveryElementsAfterSecondUnregister.Count == 0);

            // Third unregister
            Assert.IsTrue(discoveryElementsAfterThirdUnregister.Count == 0);

            // First update
            Assert.IsNotNull(applicationIdFromFirstUpdate);
            Assert.AreNotEqual(applicationIdFromFirstRegister, applicationIdFromFirstUpdate);
            AfterGdsRegistrationAsserts(discoveryElementsAfterFirstUpdate, applicationElement);
            
            // Second register
            Assert.IsNotNull(applicationIdFromSecondRegister);
            Assert.AreNotEqual(applicationIdFromFirstRegister, applicationIdFromSecondRegister);
            Assert.AreNotEqual(applicationIdFromFirstUpdate, applicationIdFromSecondRegister);
            AfterGdsRegistrationAsserts(discoveryElementsAfterSecondRegister, applicationElement);

            // Second update
            Assert.IsNotNull(applicationIdFromSecondUpdate);
            Assert.AreEqual(applicationIdFromSecondRegister, applicationIdFromSecondUpdate);
            AfterGdsRegistrationAsserts(discoveryElementsAfterSecondUpdate, applicationElement);
        }

        [TestMethod]
        public void ObtainNewCertificateTest()
        {
            TestSection.Arrange();
            var gdsEndpointDescriptor = new UAEndpointDescriptor(UATestEnvironment.OpcFoundation_GlobalDiscoveryServer.EndpointDescriptors.OpcTcp)
                { UserIdentity = OpcFoundation_GlobalDiscoveryServer.UserIdentities.Admin };
            AssertUtilities.WriteLine($"gdsEndpointDescriptor: {gdsEndpointDescriptor}");
            TestEnvironment.PrecheckEndpoint(gdsEndpointDescriptor);

            var client = new EasyUAClient();
            IEasyUAClientApplication application = client.GetService<IEasyUAClientApplication>();
            Check.AssertNotNull(application);
            string certificateSubjectName = application.GetCertificateSubjectName();

            TestSection.Act();
            PkiCertificate certificate = application.ObtainNewCertificate(gdsEndpointDescriptor,
                new Progress<string>(s => AssertUtilities.WriteLine($"Progress: {s}")));

            TestSection.Assert();
            Assert.IsNotNull(certificate);
            AssertUtilities.WriteLine(certificate);
            Assert.AreEqual(certificateSubjectName, certificate.Subject);
        }

        [DynamicData(nameof(UATestEnvironment.OpcFoundation_GlobalDiscoveryServer_AllEndpointDescriptorDataRows), typeof(UATestEnvironment))]
        [TestMethod]
        public void RefreshTrustListsTest([NotNull] UAEndpointDescriptor endpointDescriptor)
        {
            TestSection.Arrange();
            var gdsEndpointDescriptor = new UAEndpointDescriptor(endpointDescriptor)
                { UserIdentity = OpcFoundation_GlobalDiscoveryServer.UserIdentities.Admin };
            AssertUtilities.WriteLine($"gdsEndpointDescriptor: {gdsEndpointDescriptor}");
            TestEnvironment.PrecheckEndpoint(gdsEndpointDescriptor);

            var client = new EasyUAClient();
            IEasyUAClientApplication application = client.GetService<IEasyUAClientApplication>();
            Check.AssertNotNull(application);

            TestSection.Act();
            UATrustListMasks masks = application.RefreshTrustLists(gdsEndpointDescriptor);

            TestSection.Assert();
            AssertUtilities.WriteLine(masks);
        }

        [TestMethod]
        public void RegisterToGdsTest()
        {
            TestSection.Arrange();
            var gdsEndpointDescriptor = new UAEndpointDescriptor(UATestEnvironment.OpcFoundation_GlobalDiscoveryServer.EndpointDescriptors.OpcTcp)
                { UserIdentity = OpcFoundation_GlobalDiscoveryServer.UserIdentities.Admin };
            AssertUtilities.WriteLine($"gdsEndpointDescriptor: {gdsEndpointDescriptor}");
            TestEnvironment.PrecheckEndpoint(gdsEndpointDescriptor);

            var client = new EasyUAClient();
            IEasyUAClientApplication application = client.GetService<IEasyUAClientApplication>();
            Check.AssertNotNull(application);

            TestSection.Act();
            UANodeId applicationId = application.RegisterToGds(gdsEndpointDescriptor);

            TestSection.Assert();
            Assert.IsNotNull(applicationId);
        }

        [TestMethod]
        public void UnregisterFromGdsTest()
        {
            TestSection.Arrange();
            var gdsEndpointDescriptor = new UAEndpointDescriptor(UATestEnvironment.OpcFoundation_GlobalDiscoveryServer.EndpointDescriptors.OpcTcp)
                { UserIdentity = OpcFoundation_GlobalDiscoveryServer.UserIdentities.Admin };
            AssertUtilities.WriteLine($"gdsEndpointDescriptor: {gdsEndpointDescriptor}");
            TestEnvironment.PrecheckEndpoint(gdsEndpointDescriptor);

            var client = new EasyUAClient();
            IEasyUAClientApplication application = client.GetService<IEasyUAClientApplication>();
            Check.AssertNotNull(application);

            TestSection.Act();
            application.UnregisterFromGds(gdsEndpointDescriptor);

            TestSection.Assert();
        }

        [TestMethod]
        public void UpdateGdsRegistrationTest()
        {
            TestSection.Arrange();
            var gdsEndpointDescriptor = new UAEndpointDescriptor(UATestEnvironment.OpcFoundation_GlobalDiscoveryServer.EndpointDescriptors.OpcTcp)
                { UserIdentity = OpcFoundation_GlobalDiscoveryServer.UserIdentities.Admin };
            AssertUtilities.WriteLine($"gdsEndpointDescriptor: {gdsEndpointDescriptor}");
            TestEnvironment.PrecheckEndpoint(gdsEndpointDescriptor);

            var client = new EasyUAClient();
            IEasyUAClientApplication application = client.GetService<IEasyUAClientApplication>();
            Check.AssertNotNull(application);

            TestSection.Act();
            UANodeId applicationId = application.UpdateGdsRegistration(gdsEndpointDescriptor);

            TestSection.Assert();
            Assert.IsNotNull(applicationId);
        }

        #endregion



        #region Private Methods

        private void AfterGdsRegistrationAsserts(
            [NotNull] UADiscoveryElementCollection discoveryElementsAfterRegister,
            [NotNull] UAApplicationElement applicationElement)
        {
            Assert.AreEqual(1, discoveryElementsAfterRegister.Count);
            UADiscoveryElement discoveryElement = discoveryElementsAfterRegister[0];
            Check.AssertNotNull(discoveryElement);
            AssertUtilities.WriteLine($"discoveryElement: {discoveryElement}");
            Assert.AreEqual(discoveryElement.ApplicationName, applicationElement.ApplicationName);
            Assert.AreEqual(discoveryElement.ApplicationType, applicationElement.ApplicationType);
            Assert.AreEqual(discoveryElement.ApplicationUriString, applicationElement.ApplicationUriString);
            Assert.IsTrue(new HashSet<string>(discoveryElement.DiscoveryUriStrings).SetEquals(new HashSet<string>(applicationElement.DiscoveryUriStrings)));
            Assert.AreEqual(discoveryElement.ProductUriString, applicationElement.ProductUriString);
            Assert.IsTrue(new HashSet<string>(discoveryElement.ServerCapabilities).SetEquals(new HashSet<string>(applicationElement.ServerCapabilities)));
        }

        #endregion



        #region Private Properties

        [NotNull]
        private UATestEnvironment TestEnvironment => UATestEnvironment.Current;

        #endregion
    }
}
