﻿using System;
using System.Collections;
using JetBrains.Annotations;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using OpcLabs.BaseLib.DataTypeModel;
using OpcLabs.EasyOpc.UA.DataTypeModel;
using OpcLabs.EasyOpc.UA.DataTypeModel.Implementation;

namespace OpcLabs.EasyOpcTest.UA.DataTypeModel.Internal
{
    partial class TypeDescriptionExamplesTest
    {
        #region Static Private Properties

        [NotNull]
        static private StructuredDataType QualityDataType
        {
            get
            {
                var structuredDataType = new StructuredDataType("Quality", "TypeDescriptionExamples:Quality")
                {
                    new DataField("LimitBits", new SequenceDataType(2, UAOpcBinaryStandardDataTypes.Bit)),
                    new DataField("QualityBits", new SequenceDataType(6, UAOpcBinaryStandardDataTypes.Bit)),
                    new DataField("VendorBits", UAOpcBinaryStandardDataTypes.Byte)
                };
                structuredDataType.Description = "An OPC COM-DA quality value.";
                return structuredDataType;
            }
        }

        [NotNull] static private StructuredData Quality_0Data => CreateQualityData(0, 0, 0);
        [NotNull] static private byte[] Quality_0ExtensionObjectBody => new byte[2];

        [NotNull] static private StructuredData Quality_1Data => CreateQualityData(1, 0, 0);
        [NotNull] static private byte[] Quality_1ExtensionObjectBody => new byte[] { 1, 0 };

        [NotNull] static private StructuredData Quality_2Data => CreateQualityData(2, 0, 0);
        [NotNull] static private byte[] Quality_2ExtensionObjectBody => new byte[] { 2, 0 };

        [NotNull] static private StructuredData Quality_192Data => CreateQualityData(0, 48, 0);
        [NotNull] static private byte[] Quality_192ExtensionObjectBody => new byte[] { 192, 0 };

        [NotNull] static private StructuredData Quality_256Data => CreateQualityData(0, 0, 1);
        [NotNull] static private byte[] Quality_256ExtensionObjectBody => new byte[] { 0, 1 };

        [NotNull] static private StructuredData Quality_32768Data => CreateQualityData(0, 0, 128);
        [NotNull] static private byte[] Quality_32768ExtensionObjectBody => new byte[] { 0, 128 };

        #endregion



        #region Static Private Methods

        [NotNull]
        static private StructuredData CreateQualityData(byte limitBits, byte qualityBits, byte vendorBits) =>
            new StructuredData(QualityDataType)
            {
                {"LimitBits", SequenceData.BitsFromValue(UAOpcBinaryStandardDataTypes.Bit, value:limitBits, bitCount:2)},
                {"QualityBits", SequenceData.BitsFromValue(UAOpcBinaryStandardDataTypes.Bit, value:qualityBits, bitCount:6)},
                {"VendorBits", new PrimitiveData(UAOpcBinaryStandardDataTypes.Byte, vendorBits) }
            };

        #endregion



        #region Test Methods

        [TestMethod]
        public void DecodeTest_Quality_0() =>
            DecodeTest(null, TypeDictionaryByteArray, "Quality", Quality_0ExtensionObjectBody, Quality_0Data);

        [TestMethod]
        public void DecodeTest_Quality_1() =>
            DecodeTest(null, TypeDictionaryByteArray, "Quality", Quality_1ExtensionObjectBody, Quality_1Data);

        [TestMethod]
        public void DecodeTest_Quality_2() =>
            DecodeTest(null, TypeDictionaryByteArray, "Quality", Quality_2ExtensionObjectBody, Quality_2Data);

        [TestMethod]
        public void DecodeTest_Quality_192() =>
            DecodeTest(null, TypeDictionaryByteArray, "Quality", Quality_192ExtensionObjectBody, Quality_192Data);

        [TestMethod]
        public void DecodeTest_Quality_256() =>
            DecodeTest(null, TypeDictionaryByteArray, "Quality", Quality_256ExtensionObjectBody, Quality_256Data);

        [TestMethod]
        public void DecodeTest_Quality_32768() =>
            DecodeTest(null, TypeDictionaryByteArray, "Quality", Quality_32768ExtensionObjectBody, Quality_32768Data);

        [TestMethod]
        public void EncodeErrorTest_BinaryEncoderStructuredDataFieldNotPresent() =>
            EncodeErrorTest(null, 
                TypeDictionaryByteArray,
                "Quality",
                new StructuredData(),
                UADataTypeModelErrorCodes.OpcBinaryEncoderStructuredDataFieldNotPresent);

        [TestMethod]
        public void EncodeErrorTest_BinaryEncoderSequenceFieldNotSequenceData()
        {
            var structuredData = new StructuredData { { "LimitBits", new OpaqueData(new BitArray(128)) } };
            EncodeErrorTest(null, 
                TypeDictionaryByteArray,
                "Quality",
                structuredData,
                UADataTypeModelErrorCodes.OpcBinaryEncoderSequenceFieldNotSequenceData);
        }

        [TestMethod]
        public void EncodeErrorTest_BinaryEncoderSequenceDataActualLengthMismatch()
        {
            var structuredData = new StructuredData
            {
                { "LimitBits", new SequenceData
                    {
                        new PrimitiveData(false)
                    }}
            };
            EncodeErrorTest(null, 
                TypeDictionaryByteArray,
                "Quality",
                structuredData,
                UADataTypeModelErrorCodes.OpcBinaryEncoderSequenceDataActualLengthMismatch);
        }

        [TestMethod]
        public void EncodeErrorTest_BinaryEncoderPrimitiveStandardDataTypeMismatch()
        {
            var structuredData = new StructuredData
            {
                {"LimitBits", new SequenceData
                {
                    new PrimitiveData(false),
                    new PrimitiveData(false)
                }},
                {"QualityBits", new SequenceData
                {
                    new PrimitiveData(false),
                    new PrimitiveData(false),
                    new PrimitiveData(false),
                    new PrimitiveData(false),
                    new PrimitiveData(false),
                    new PrimitiveData(false)
                }},
                {"VendorBits", new PrimitiveData((UInt16)0) }
            };
            EncodeErrorTest(
                // ReSharper disable once PossibleNullReferenceException
                dts => FindStandardBidtd(dts).AllowConvertiblePrimitiveDataValues = false,
                TypeDictionaryByteArray,
                "Quality",
                structuredData,
                UADataTypeModelErrorCodes.StandardDictionaryPrimitiveDataTypeMismatch);
        }

        [TestMethod]
        public void EncodeTest_Quality_0() =>
            EncodeTest(null, TypeDictionaryByteArray, "Quality", Quality_0Data, Quality_0ExtensionObjectBody);

        [TestMethod]
        public void EncodeTest_Quality_1() =>
            EncodeTest(null, TypeDictionaryByteArray, "Quality", Quality_1Data, Quality_1ExtensionObjectBody);

        [TestMethod]
        public void EncodeTest_Quality_2() =>
            EncodeTest(null, TypeDictionaryByteArray, "Quality", Quality_2Data, Quality_2ExtensionObjectBody);

        [TestMethod]
        public void EncodeTest_Quality_192() =>
            EncodeTest(null, TypeDictionaryByteArray, "Quality", Quality_192Data, Quality_192ExtensionObjectBody);

        [TestMethod]
        public void EncodeTest_Quality_256() =>
            EncodeTest(null, TypeDictionaryByteArray, "Quality", Quality_256Data, Quality_256ExtensionObjectBody);

        [TestMethod]
        public void EncodeTest_Quality_32768() =>
            EncodeTest(null, TypeDictionaryByteArray, "Quality", Quality_32768Data, Quality_32768ExtensionObjectBody);

        #endregion
    }
}
